
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "heap.h"
#include "runtime.h"


#define HEAPSIZE 0x100000  /* 1 MB */

/* current heap */
scalar * currStart;
scalar * allocPtr;
scalar * allocLimit;

/* alternate heap */
scalar * altStart;
scalar * altLimit;


void gc (scalar *roots, unsigned rootnum);


scalar alloc (unsigned size, scalar flags, scalar *roots, unsigned rootnum)
{
   scalar *ptr;
   unsigned sizeAlloc = size+1;

   if (allocPtr + sizeAlloc > allocLimit)
      {
      fprintf(stderr, "[collecting");
      fflush(stdout);

      gc (roots, rootnum);

      fprintf(stderr, "]");
      fflush(stderr);

      if (allocPtr + sizeAlloc > allocLimit)
	 {
	 fprintf (stderr, "Out of memory.\n");
	 exit (1);
	 }
      }
   
   ptr = allocPtr;
   allocPtr += sizeAlloc;

   ptr[0] = (size << 2) | HEAPOBJ | flags;
   memset (ptr+1, 0, size*4);

   return ((scalar)ptr);
}


void halt (void)
{
   free (currStart);
   free (altStart);
   exit (0);
}


scalar intToString (scalar n, scalar *roots, unsigned rootnum)
{
   scalar x = ston(n);
   unsigned sz = 0;
   char *p, *pcur;

   if (x == 0)
      {
      p = "@@@@0";
      return ((scalar)p);
      }

   if (x < 0)
      {
      sz = 1;
      x = -x;
      }

   while (x > 0)
      {
      sz++;
      x = x / 10;
      }

   p = (char *) alloc (sz / 4 + 1, NOTRACE, roots, rootnum);

   x = ston(n);
   pcur = p+4+sz;

   if (x < 0)
      {
      *(p+4) = '~';
      x = -x;
      }

   *(pcur--) = 0;

   while (x > 0)
      {
      *(pcur--) = '0' + (x % 10);
      x = x / 10;
      }

   return ((scalar)p);
}


scalar concat (scalar *str1, scalar *str2, scalar *roots, unsigned rootnum)
{
   unsigned sz1 = strlen(stostr(*str1));
   unsigned sz2 = strlen(stostr(*str2));

   char *p = (char *) alloc ((sz1 + sz2) / 4 + 1, NOTRACE, roots, rootnum);
   /* str1 and str2 may have moved */

   strcpy (p+4, stostr(*str1));
   strcpy (p+4+sz1, stostr(*str2));

   return ((scalar)p);
}


scalar print (scalar str, scalar *roots, unsigned rootnum)
{
   fprintf(stdout, "%s", stostr(str));
   fflush(stdout);
   return (alloc (0, 0, roots, rootnum));
}


extern void mlmain (scalar *argp, codeptr *nextfn);


int main (void)
{
   scalar arg;
   codeptr fn;

   if (sizeof(scalar) != 4 || sizeof(scalar *) != 4 || sizeof(char) != 1)
      {
      fprintf (stderr, "Unsupported architecture.\n");
      exit(-1);
      }

   /* Initialize the heap. */
   currStart = (scalar *) malloc (HEAPSIZE);
   altStart = (scalar *) malloc (HEAPSIZE);
   allocLimit = (scalar *) ((char *)currStart + HEAPSIZE);
   altLimit = (scalar *) ((char *)altStart + HEAPSIZE);
   allocPtr = currStart;

   fn = &mlmain;
   arg = 1;

   while (1)
      {
      (*fn) (&arg, &fn);
      }

}
